%% Algoritmo de abejas para generar un autoencoder
% Este programa entrena una red nueronal en configuracion de autoencoder
% Se calculan los pesos y en la capa intermedia se tienen 3 componentes
% Ademas, en esta configuracion se tiene una capa de salida para generar el
% autoencoder

clc, clear, close all;

%% Cargar los 2000 datos que corresponden a 2 tipos de vinos (tinto-blanco)
load('target_no_lineal.mat')
global TARGET
TARGET = target;

%% Parametros iniciales 
D     =   3;            % Dimension (features) ver red neuronal 3 neuronas en capa intermedias y 17 de salida
Np    = 100;            % Tamano de la poblacion de abejas
Fc    = round(Np/2);    % Fuente de comida
Gmax  = 500;            % Generaciones (Iteraciones) maximas
Limit =  30;            % Criterio de abandono
xmin  = -10;            % Liminte inferior
xmax  =  10;            % Limite superior

amin =  -1;
amax =   1;
bmin =  -1;
bmax =   1;
cmin = -99;
cmax = -10;
dmin = -10;
dmax =  10;
capmin = 10;
capmax = 150;

%% Poblacion inicial aleatoria

% Tabla Vacia de abejas 
Abeja_vacia.Posicion       = [];
Abeja_vacia.Costo          = [];
Abeja_vacia.Fitness        = [];
Abeja_vacia.Limite         = [];
Abeja_vacia.Probabi        = [];
Abeja_vacia.Mejor.Posicion = [];
Abeja_vacia.Mejor.Costo    = [];

% Creamos el arreglo de la poblacion de abejas las Abejas
Abejas = repmat(Abeja_vacia, Np, 1);

% Se guarda en memoria el mejor dato global de las abejas
Mejor_Global.Costo = inf;

i = 1;
while (i <= Np)
    aa = unifrnd(amin, amax, 1);
    bb = unifrnd(bmin, bmax, 1);
    cc = unifrnd(cmin, cmax, 1);
    dd = unifrnd(dmin, dmax, 1);
    cap = unifrnd(capmin, capmax, 1);
    Abejas(i).Posicion = [bb cc dd];
    Abejas(i).Costo    = Spiking_no_lineal( Abejas(i).Posicion );
    Abejas(i).Fitness  = Fitness(Abejas(i).Costo);
    Abejas(i).Limite   = 0;
    Abejas(i).Probabi  = 0;
    
    % Actualizar el mejor valor global
    if Abejas(i).Costo < Mejor_Global.Costo
        Mejor_Global = Abejas(i);
    end
    
    i = i + 1;
    
end

%% Inicia el bucle de las abejas

g = 1;                  % Contador de generaciones (iteraciones)
while ( g <= Gmax )
    
    %====================== Incio de face abejas obreras =================%
    a = 1;              % numero de abeja
    while (a <= Np)
        
        % Parametro de dimension a modificar aleatoriamente (cualquier feature)
        DimChange = fix( rand*D ) + 1;
        
        % Compara abeja actual con otra del enjambre
        Vecino = fix( rand*Np ) + 1;
        while( Vecino == a )
            Vecino = fix( rand*Np ) + 1;
        end
        
        V   = Abejas(a);
        phi = (rand - 0.5) * 2;
        V.Posicion(1, DimChange) = Abejas(a).Posicion(1, DimChange) + ...
               ( phi * (Abejas(a).Posicion(1, DimChange) - Abejas(Vecino).Posicion(1, DimChange)));
        
        %Si el valor del nuevo parametro esta fuera de los limites hay que restringirlo
%         ind = find( V.Posicion < xmin );
%         V.Posicion(1, ind) = xmin;
%         ind = find( V.Posicion > xmax );
%         V.Posicion(1, ind) = xmax;
        
        V.Costo            = Spiking_no_lineal( V.Posicion );
        V.Fitness          = Fitness(V.Costo);
        V.Limite           = 0;
        
        % Se toma el mejor valor
        if V.Costo < Abejas(a).Costo
            Abejas(a) = V;
        else
            Abejas(a).Limite = Abejas(a).Limite + 1;
        end
        
        % Actualizar el mejor valor global
        if Abejas(a).Costo < Mejor_Global.Costo
            Mejor_Global = Abejas(a);
        end
        
        a = a + 1;
    end
    %=====================================================================%
    %================= Incio de face abejas observadoras =================%
    % Se calcula la probabilidad utilizando los valores de fitness
    % normalizados
    i = 1;
    while (i <= Np)
        Abejas(i).Probabi = ( 0.9 * (Abejas(i).Fitness/Mejor_Global.Fitness) ) + 0.1;
        i = i + 1;
    end
    
    i=1;
    t=0;
    while(t <= Np)
        if( rand <= Abejas(i).Probabi )
            t = t+1;
            
            % Parametro de dimension a modificar aleatoriamente (cualquier feature)
            DimChange = fix( rand*D ) + 1;
            
            % Compara abeja actual con otra del enjambre
            Vecino = fix( rand*Np ) + 1;
            while( Vecino == i )
                Vecino = fix( rand*Np ) + 1;
            end
            
            V   = Abejas(i);
            phi = (rand - 0.5) * 2;
            V.Posicion(1, DimChange) = Abejas(i).Posicion(1, DimChange) + ...
                ( phi * (Abejas(i).Posicion(1, DimChange) - Abejas(Vecino).Posicion(1, DimChange)));
            
            %Si el valor del nuevo parametro esta fuera de los limites hay que restringirlo
%             ind = find( V.Posicion < xmin );
%             V.Posicion(1, ind) = xmin;
%             ind = find( V.Posicion > xmax );
%             V.Posicion(1, ind) = xmax;
            
            V.Costo            = Spiking_no_lineal( V.Posicion );
            V.Fitness          = Fitness(V.Costo);
            V.Limite           = 0;
            
            % Se toma el mejor valor
            if V.Costo < Abejas(i).Costo
                Abejas(i) = V;
            else
                Abejas(i).Limite = Abejas(i).Limite + 1;
            end
            
            % Actualizar el mejor valor global
            if Abejas(i).Costo < Mejor_Global.Costo
                Mejor_Global = Abejas(i);
            end
            
        end
        i = i+1;
        if (i == (Np)+1)
            i=1;
        end
    end
    %=====================================================================%
    %================== Incio de face abeja exploradora ==================%
    b = 1;
    while (b <= Np)
        
        if (Abejas(b).Limite >= Limit )
            aa = unifrnd( amin, amax, 1);
            bb = unifrnd( bmin, bmax, 1);
            cc = unifrnd( cmin, cmax, 1);
            dd = unifrnd( dmin, dmax, 1);
            cap = unifrnd(capmin, capmax, 1);
            Abejas(b).Posicion = [bb cc dd];
            Abejas(b).Costo    = Spiking_no_lineal( Abejas(b).Posicion );
            Abejas(b).Fitness  = Fitness(Abejas(b).Costo);
            Abejas(b).Limite   = 0;
            Abejas(b).Probabi  = 0;
            
            % Actualizar el mejor valor global
            if Abejas(b).Costo < Mejor_Global.Costo
                Mejor_Global = Abejas(b);
            end
            
        end
        b = b + 1;
    end
    %=====================================================================%
    % Imprimir informacion
    disp( ['En la iteracion ', num2str(g), ' el valor global del costo es ', num2str(Mejor_Global.Costo)] )
    
    g = g + 1;
    
end

%% Fase de prueba de resultados
par = Mejor_Global.Posicion;
% a  = par(1,1); 
aa = 0.01;
b = par(1,1); c = par(1,2); d = par(1,3);
C  = 100;
vr = -60;
vt = -40;
k = 0.7;
% a = 0.03;  b = -2;   c = -50;  d = 100;
% a = 0.02;  b = -0.1; c = -55;  d = 6;
vpeak = 35;
T = 1000;    tau = 1; % time span and step (ms) ==> 1 seg
% T = 500;    tau = 1; % time span and step (ms) ==> 1/2 seg
% T = 250;    tau = 1; % time span and step (ms) ==> 1/4 seg
n = round( T/tau );
PotMem = vr * ones(1,n);     u = 0*PotMem;
kk = 0;
noLineal = zeros(1,71);
Ie = 71:141;
for ie = Ie
    kk = kk + 1;
    I = [ zeros(1,0.01*n), ie*ones(1,0.99*n) ];
    pulsos = 0;
    for i = 1:n-1
        PotMem(i+1) = PotMem(i) + tau*(k*(PotMem(i)-vr)*(PotMem(i)-vt)-u(i)+I(i))/C;
        u(i+1) = u(i) + tau * aa * (b*(PotMem(i)-vr)-u(i) );
        if PotMem(i+1) >= vpeak
            PotMem(i) = vpeak;
            PotMem(i+1) = c;
            u(i+1) = u(i+1) + d;
            pulsos = pulsos + 1;
        end
    end
%     pulsos = pulsos*2; % medio segundo multiplicado por dos == frec(hz)
    noLineal(kk) = pulsos;
end
figure, plot(Ie, noLineal, '.-r')

%% Esta funcion permite calcular el fitness relativo
function Out = Fitness(In)
    if In > 0
        Out = 1 / (1 + In);
    else
        Out = 1 + abs( In );
    end
end

%% Funcin de optimizacin para el autoencoder, compara entrada vs salida
function Out = Spiking_no_lineal(Parametros)
global TARGET
% a  = Parametros(1,1); 
b = Parametros(1,1); c = Parametros(1,2); d = Parametros(1,3);
a = 0.01;
C  = 100;
vr = -60;
vt = -40;
k  = 0.7;
% a = 0.03;  b = -2;   c = -50;  d = 100;
% a = 0.02;  b = -0.1; c = -55;  d = 6;
vpeak = 35;
T = 1000;    tau = 1; % time span and step (ms) ==> 1 seg
% T = 500;    tau = 1; % time span and step (ms) ==> 1/2 seg
% T = 250;    tau = 1; % time span and step (ms) ==> 1/4 seg
n = round( T/tau );
PotMem = vr * ones(1,n);     u = 0*PotMem;
kk = 0;
noLineal = zeros(1,71);
for Ie = 71:141
    kk = kk + 1;
    I = [ zeros(1,0.01*n), Ie*ones(1,0.99*n) ];
    pulsos = 0;
    for i = 1:n-1
        PotMem(i+1) = PotMem(i) + tau*(k*(PotMem(i)-vr)*(PotMem(i)-vt)-u(i)+I(i))/C;
        u(i+1) = u(i) + tau * a * (b*(PotMem(i)-vr)-u(i) );
        if PotMem(i+1) >= vpeak
            PotMem(i) = vpeak;
            PotMem(i+1) = c;
            u(i+1) = u(i+1) + d;
            pulsos = pulsos + 1;
        end
    end
%     pulsos = pulsos*2; % medio segundo multiplicado por dos == frec(hz)
%     pulsos = pulsos*4; % un cuarto de segundo multiplicado por cuatro == frec(hz)
    noLineal(kk) = pulsos;
end
Out = immse(TARGET, noLineal);

end


